import logging
log = logging.getLogger(__name__)

from collections.abc import Iterable
import threading

import numpy as np
import sounddevice as sd
sd.default.latency = 'low'


class BaseCallbackContext:

    def __init__(self, cb=None):
        self.i = 0
        if cb is None:
            cb = lambda x: x
        self.cb = cb

    def _valid_samples(self, samples, status):
        if status:
            log.warning('portaudio callback status: %r', status)

        # Calculate the number of valid samples remaining
        samples_remaining = self.n - self.i
        log.info('%r: %r', self.n, self.i)
        if samples_remaining == 0:
            raise sd.CallbackStop

        valid_samples = min(samples_remaining, samples)
        result = self.i, valid_samples
        self.i += valid_samples
        self.cb(valid_samples)
        return result

    def __call__(self, *args):
        raise NotImplementedError


class RecordCallbackContext(BaseCallbackContext):

    def __init__(self, input_buffer, input_scale, cb=None):
        super().__init__(cb=cb)
        self.input_buffer = input_buffer
        self.input_scale = input_scale
        self.n = len(input_buffer)

    def __call__(self, indata, samples, time, status):
        # Read the next segment to the input buffer
        i, valid_samples = self._valid_samples(samples, status)
        self.input_buffer[i:i + valid_samples] = indata[:valid_samples] / self.input_scale


class QueueCallbackContext:

    def __init__(self, queues):
        super().__init__()
        self.queues = queues

    def __call__(self, outdata, samples, time, status):
        if status:
            log.warning('portaudio callback status: %r', status)
        for i, q in enumerate(self.queues):
            outdata[:, i] = q.pop_buffer(samples)

        # Check if any queues are empty
        for q in self.queues:
            if not q._empty:
                break
        else:
            raise sd.CallbackStop


class PlayCallbackContext(BaseCallbackContext):

    def __init__(self, output_buffer, cb=None):
        super().__init__(cb=cb)
        self.output_buffer = output_buffer
        self.n = len(output_buffer)

    def __call__(self, outdata, samples, time, status):
        # Write the next segment to the output buffer
        i, valid_samples = self._valid_samples(samples, status)
        outdata[:valid_samples] = self.output_buffer[i:i + valid_samples]
        outdata[valid_samples:] = 0


class PlayRecordCallbackContext(BaseCallbackContext):

    def __init__(self, input_buffer, output_buffer, input_scale, cb=None):
        super().__init__(cb=cb)
        self.input_buffer = input_buffer
        self.output_buffer = output_buffer
        self.input_scale = input_scale
        self.n = len(output_buffer)

    def __call__(self, indata,  outdata, samples, time, status):
        # Read/write the next segments
        i, valid_samples = self._valid_samples(samples, status)
        outdata[:valid_samples] = self.output_buffer[i:i + valid_samples]
        outdata[valid_samples:] = 0
        self.input_buffer[i:i + valid_samples] = indata[:valid_samples] / self.input_scale


class SoundDevice:
    '''
    Basic wrapper around an audio device that provides play and record
    functionality.

    Parameters
    ----------
    input_device : string
        Name of input device as seen by portaudio.
    output_device : string
        Name of output device as seen by portaudio.
    input_scale : float
        Scaling factor to apply to input stream. This is primarily used by the
        ASIO Fireface interface for the Babyface since the input data needs to
        be scaled by ~1/3 to match the expected amplitude.
    '''

    def __init__(self, input_device, output_device, input_scale=1, fs=None):
        self.input_device = input_device
        self.input_info = sd.query_devices(self.input_device)
        self.input_scale = input_scale
        log.info('Properties for input device %r: %r', self.input_device, self.input_info)

        self.output_device = output_device
        self.output_info = sd.query_devices(self.output_device)
        log.info('Properties for output device %r: %r', self.output_device, self.output_info)

        if fs is None:
            self.fs = self.input_info['default_samplerate']
        else:
            self.fs = fs

    def _start_stream(self, stream_class, **stream_kw):
        self.event = threading.Event()
        return stream_class(finished_callback=self.event.set, **stream_kw)

    def wait(self, timeout=0.1):
        return self.event.wait(timeout)

    def join(self):
        while True:
            if self.event.wait(0.1):
                break

    def play_async_queue(self, queues, output_channels=None, cb=None):
        output_settings = None
        input_settings = None
        if output_channels is None:
            output_channels = len(queues)
        elif isinstance(output_channels, Iterable):
            if len(output_channels) != len(queues):
                m = 'Output mapping of channels does not match waveform shape'
                raise ValueError(m)
            output_settings = sd.AsioSettings(output_channels)
            output_channels = len(output_channels)
        return self._start_stream(
            sd.OutputStream,
            device=self.output_device,
            samplerate=self.fs,
            blocksize=1024,
            callback=QueueCallbackContext(queues),
            channels=output_channels,
            extra_settings=output_settings,
        )

    def play_queue(self, queues, output_channels=None):
        with self.play_async_queue(queues, output_channels) as stream:
            self.join()

    def play_async(self, waveform, output_channels=None, cb=None):
        if waveform.ndim == 1:
            waveform = waveform[np.newaxis]
        output_settings = None
        input_settings = None
        if output_channels is None:
            output_channels = len(waveform)
        elif isinstance(output_channels, Iterable):
            if len(output_channels) != len(waveform):
                m = 'Output mapping of channels does not match waveform shape'
                raise ValueError(m)
            output_settings = sd.AsioSettings(output_channels)
            output_channels = len(output_channels)

        return self._start_stream(
            sd.OutputStream,
            device=self.output_device,
            samplerate=self.fs,
            blocksize=1024,
            callback=PlayCallbackContext(waveform.T, cb=cb),
            channels=output_channels,
            extra_settings=output_settings,
        )

    def play(self, waveform, output_channels=None, cb=None):
        with self.play_async(waveform, output_channels, cb=cb) as stream:
            self.join()

    def acquire(self, waveform, input_channels=1, output_channels=None):
        if waveform.ndim == 1:
            waveform = waveform[np.newaxis]
        output_settings = None
        input_settings = None
        if output_channels is None:
            output_channels = len(waveform)
        elif isinstance(output_channels, Iterable):
            if len(output_channels) != len(waveform):
                m = 'Output mapping of channels does not match waveform shape'
                raise ValueError(m)
            output_settings = sd.AsioSettings(output_channels)
            output_channels = len(output_channels)
        if isinstance(input_channels, Iterable):
            input_settings = sd.AsioSettings(input_channels)
            input_channels = len(input_channels)

        recording = np.zeros((waveform.shape[-1], input_channels), dtype='double')
        stream = self._start_stream(
            sd.Stream,
            device=(self.input_device, self.output_device),
            samplerate=self.fs,
            blocksize=1024,
            callback=PlayRecordCallbackContext(recording, waveform.T, self.input_scale),
            channels=(input_channels, output_channels),
            extra_settings=(input_settings, output_settings),
        )
        with stream:
            self.join()
        return recording.T

    def record(self, n_samples, input_channels):
        input_settings = None
        if isinstance(input_channels, Iterable):
            input_settings = sd.AsioSettings(input_channels)
            input_channels = len(input_channels)

        recording = np.zeros((n_samples, input_channels))
        stream = self._start_stream(
            sd.InputStream,
            device=self.input_device,
            samplerate=self.fs,
            blocksize=1024,
            callback=RecordCallbackContext(recording, self.input_scale),
            channels=input_channels,
            extra_settings=input_settings,
        )
        with stream:
            self.join()
        return recording.T



from pathlib import Path
from psiaudio.stim import load_wav

import logging
logging.basicConfig(level='INFO')

import matplotlib.pyplot as plt
from scipy import signal
import numpy as np

from psiaudio.calibration import FlatCalibration
from psiaudio.stim import ramped_tone
from psiaudio import util

from ncrar_audio.babyface import Babyface

base_path = Path(r'C:\Users\biosemi\Desktop\Koerner_CDA2')
wav_path = base_path / 'tone1k_150ms.wav'


def main():
    calibration = FlatCalibration.unity()
    device = Babyface(output_channels='earphones', trigger_channels='XLR1',
                      use_osc=False)

    wav = load_wav(device.fs, wav_path)
    wav = wav[:, :2].T

    for i in range(10):
        with device.play(wav):
            device.join()


if __name__ == '__main__':
    main()



import matplotlib.pyplot as plt
from psiaudio.queue import FIFOSignalQueue
from psiaudio.stim import load_wav
from ncrar_audio.babyface import Babyface


#device = Babyface()
fs = 44e3

wav = load_wav(fs, 'stim/da_40ms.wav')
queue = FIFOSignalQueue(fs=fs)
queue.append(wav, 10, delays=1)

b = queue.pop_buffer(int(fs*10))
plt.plot(b)
plt.show()



import logging
logging.basicConfig(level='INFO')

import matplotlib.pyplot as plt
from scipy import signal
import numpy as np

from psiaudio.calibration import FlatCalibration
from psiaudio.stim import ramped_tone
from psiaudio import util

from ncrar_audio.babyface import Babyface


def main():
    calibration = FlatCalibration.unity()
    device = Babyface('earphones', 'XLR1')
    frequencies = np.array([1e3, 2e3])[:, np.newaxis]
    tone = ramped_tone(device.fs, frequency=frequencies, duration=0.5,
                       rise_time=2.5e-3, level=-16.7, calibration=calibration)

    #device.set_output('XLR')
    #device.play(tone)
    for i in range(10):
        device.play(tone)

    return
    recording = device.play(tone, 2, [0, 3])
    psd = util.db(util.psd_df(recording, device.fs)).iloc[:, 1:]

    plt.plot(psd.columns, psd.iloc[0].values, label='In 1')
    plt.plot(psd.columns, psd.iloc[1].values, label='In 2')
    plt.xlabel('Frequency (Hz)')
    plt.ylabel('Signal (dBVrms)')
    plt.show()


if __name__ == '__main__':
    main()



from .sound_device import SoundDevice


class Dante(SoundDevice):

    def __init__(self):
        super().__init__('Dante Virtual Soundcard (x64)',
                         'Dante Virtual Soundcard (x64)', fs=48e3)

    def play_azimuth(self, signal, azimuth):
        pass



import numpy as np
from scipy import signal


def _square_waveform(fs, iti=0.02, duration=0.01, repeat=1):
    n_samples = int(round(fs * iti))
    n_trig = int(round(fs * duration))

    trig = np.zeros(n_samples)
    trig[:n_trig] = 1
    trig = np.tile(trig, (repeat, 1))
    trig[1:] *= 0.5
    trig = trig.ravel()
    return trig


def _cos_waveform(fs, pulse_frequency=128, pulse_cycles=1, repeat=1):
    n_samp = int(round((1 / pulse_frequency) * fs)) * pulse_cycles
    t = np.arange(n_samp) / fs
    y = np.sin(2 * np.pi * pulse_frequency * t)
    y = np.tile(y, (repeat, 1))
    y[1:] *= 0.5
    return y.ravel()


def make_trigger(fs, n, shape='square', shape_settings=None):
    if shape_settings is None:
        shape_settings = {}
    fn = globals()[f'_{shape}_waveform']
    trig = fn(fs, **shape_settings)
    if len(trig) > n:
        n_extra = len(trig) - n
        raise ValueError(f'Trigger sequence is too long by {n_extra} samples')

    waveform = np.zeros(n)
    waveform[:len(trig)] = trig
    return waveform


def group_triggers(x, window=10):
    '''
    Group triggers if they fall within `window` of the previous trigger.

    Parameters
    ----------
    x : array
        Array of timestamps (e.g., in units of sample number or seconds).
    window : float
        Group triggers occuring within `window` of the previous trigger. The
        timestamp for the set of triggers is the timestamp of the first trigger
        in the train.

    Returns
    -------
    trigger_times : dict
        Grouped trigger times. Dictionary keys are the number of triggers in
        each group and values are a list of times where that trigger group was
        detected. The reported times are the time of the first trigger in the
        group.

    Note
    ----
    The units of `window` and `trigger_times` are the same units of `x` (e.g.,
    if `x` is specified as trigger timestamps in seconds, then `window` would
    be seconds; if `x` is specified as trigger timestamps in samples, then
    `window` would be number of samples).
    '''
    n_trig = 1

    # This handles the edge condition to make sure the very last set of
    # triggers gets included.
    x = np.pad(x, (0, 1), constant_values=x[-1] + window * 2)

    triggers = {}
    for i in range(1, len(x)):
        if (x[i] - x[i-1]) < window:
            n_trig += 1
        else:
            triggers.setdefault(n_trig, []).append(x[i-n_trig])
            n_trig = 1

    # Convert trigger lists to arrays
    return {k: np.array(v) for k, v in triggers.items()}


def extract_triggers(x, threshold=None, group_window=10, fs=None):
    '''
    Extract and group triggers generated by the trigger interface of subclasses
    of SoundDevice that support including triggers (e.g., Babyface).

    Parameters
    ----------
    x : array
        The waveform of the trigger channel (e.g., Erg2 in the Biosemi BDF
        file).
    threshold : {None, float}
        Threshold used to detect triggers. The trigger times will be the point
        at which the leading edge of the trigger crosses threshold. If None,
        will automatically be set to 20% of the maximum value in the signal.
    group_window : float
        Group triggers occuring within `window` of the previous trigger. The
        timestamp for the set of triggers is the timestamp of the first trigger
        in the train. If `fs` is not None, `group_window` should be in seconds.
        If `fs` is None, `group_window` should be in units of samples.
    fs : {None, float}
        Sampling rate of `x`.

    Returns
    -------
    trigger_times : dict
        Grouped trigger times. Dictionary keys are the number of triggers in
        each group and values are a list of times where that trigger group was
        detected. The reported times are the time of the first trigger in the
        group. Trigger times will be in units of seconds when `fs` is provided,
        samples otherwise.
    '''
    if threshold is None:
        threshold = x.max() * 0.2

    # Find all points in x that exceed the threshold value. The default return
    # type of a logical comparision is np.bool, but we need np.int so that we
    # can properly execute the next step.
    m = (x >= threshold).astype('int')

    # Compute the difference between adjacent samples. A value of 1 indicates
    # the rising edge of the trigger and -1 the falling edge (if m had been
    # kept as a boolean array, the value for both 1 and -1 would just be True).
    d = np.diff(m)

    # Pull out the indices where the difference is 1, indicating the rising
    # (leading) edge of the trigger.
    i = np.flatnonzero(d == 1)

    if fs is not None:
        i = i / fs
    return group_triggers(i, group_window)



import pyxid2


class CodeManager:

    def __init__(self, cpod):
        self.cpod = cpod

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.cpod.clear_code()


class CPod:

    def __init__(self, pulse_duration=0):
        # Get a list of all attached XID devices then use the first one. This
        # assumes that we have only one XID device (i.e., the Cedrus C-Pod)
        # attached.
        self.dev = pyxid2.get_xid_devices()[0]
        self.dev.reset_base_timer()
        self.dev.reset_rt_timer()
        self.dev.set_pulse_duration(int(pulse_duration * 1e3))

    def set_code(self, code):
        bitmask = (code << 1) | 1
        self.dev.set_lines(bitmask)
        return CodeManager(self)

    def clear_code(self):
        self.dev.set_lines(0)



import logging
log = logging.getLogger(__name__)

import json
from pathlib import Path
import re
from threading import Lock

import numpy as np
from scipy.interpolate import interp1d

from .osc_client import OSCClient
from .sound_device import SoundDevice
from . import triggers

import sounddevice as sd

P_VOLUME = re.compile(r'/1/volume(\d+)Val')
P_MICGAIN = re.compile(r'/1/micgain(\d+)Val')


def load_volume_map():
    VOLUME_MAP_FILE = Path(__file__).parent / 'totalmix_fx_volume_map.json'
    VOLUME_MAP = json.loads(VOLUME_MAP_FILE.read_text())
    scale = np.fromiter(VOLUME_MAP.keys(), 'float32')
    db = np.fromiter(VOLUME_MAP.values(), 'float32')
    return interp1d(db, scale, 'linear')


class Babyface(SoundDevice):

    output_map = {
        'XLR': [0, 1],
        'XLR1': [0],
        'XLR2': [1],
        'earphones': [2, 3],
        'earphones_left': [2],
        'earphones_right': [3],
        None: [],
    }

    def __init__(self, output_channels='earphones', trigger_channels=None,
                 ip_address=None, send_port=7001, recv_port=9001,
                 use_osc=True):
        self._volume_db = {}
        self._mic_gain_db = {}
        self._lock = Lock()
        self._volume_map = load_volume_map()

        if use_osc:
            self.osc_client = OSCClient(ip_address, send_port, recv_port)
            self.osc_client.dispatch.map('/1/volume*Val', self._volume_updated)
            self.osc_client.dispatch.map('/1/micgain*Val', self._mic_gain_updated)

        self._output_channels = []
        self._trigger_channels = []
        self.set_output(output_channels)
        self.set_trigger(trigger_channels)

        name = 'ASIO Fireface USB'
        super().__init__(name, name, input_scale=0.3395)

    def set_output(self, output_channels):
        if isinstance(output_channels, str):
            output_channels = [output_channels]

        self._output_channels = []
        for oc in output_channels:
            self._output_channels.extend(self.output_map[oc])
        self._output_map = self._output_channels + self._trigger_channels

    def set_trigger(self, trigger_channels):
        self._trigger_channels = self.output_map[trigger_channels]
        self._output_map = self._output_channels + self._trigger_channels

    def add_trigger(self, waveform):
        trigger = triggers.make_trigger(self.fs, waveform.shape[-1])
        trigger = np.repeat(trigger[np.newaxis],
                            len(self._trigger_channels), 0)
        return np.vstack((waveform, trigger))

    def play(self, waveform, cb=None):
        if self._trigger_channels:
            waveform = self.add_trigger(waveform)
        return super().play(waveform, self._output_map, cb=cb)

    def play_queue(self, queues):
        return super().play_queue(queues, self._output_map)

    def acquire(self, waveform, input_channels):
        if self._trigger_channels:
            waveform = self.add_trigger(waveform)
        return super().acquire(waveform, input_channels, self._output_map)

    def play_mono(self, waveform, side):
        if waveform.ndim != 1:
            raise ValueError('Only a single waveform can be provided')
        if len(self._output_channels) == 1:
            self.play(waveform)
        silence = np.zeros_like(waveform)
        if side == 'left':
            waveform = np.vstack((waveform, silence))
        elif side == 'right':
            waveform = np.vstack((silence, waveform))
        else:
            raise ValueError('side must be "right" or "left"')
        self.play(waveform)

    def play_stereo(self, waveform):
        if waveform.ndim != 1:
            raise ValueError('Only a single waveform can be provided')
        if len(self._output_channels) == 1:
            raise ValueError('Only a single output channel available')
        waveform = np.vstack((waveform, waveform))
        self.play(waveform)

    def set_volume(self, db, channels=None):
        if channels is None:
            channels = np.arange(12) + 1
        volume = float(self._volume_map(db))
        messages = [(f'/1/volume{i}', volume) for i in channels]
        self.osc_client.send_messages(messages)

    def set_mic_gain(self, db, channels=None):
        if channels is None:
            channels = np.arange(12) + 1
        volume = float(self._volume_map(db))
        messages = [(f'/1/micgain{i}', volume) for i in channels]
        self.osc_client.send_messages(messages)

    def _volume_updated(self, address, value):
        with self._lock:
            channel = int(P_VOLUME.match(address).group(1))
            if value == '-oo':
                value = -np.inf
            else:
                value = float(value.split(' ')[0])
            self._volume_db[channel] = value

    def _mic_gain_updated(self, address, value):
        with self._lock:
            channel = int(P_MICGAIN.match(address).group(1))
            if value == '-oo':
                value = -np.inf
            else:
                value = float(value.split(' ')[0])
            self._mic_gain_db[channel] = value



'''
Record the output of XLR1 via the input of XLR1 on the Babyface. Connect the
input to the output with a XLR cable before running this demo.
'''

import matplotlib.pyplot as plt
import numpy as np

from psiaudio.calibration import FlatCalibration
from psiaudio.stim import tone
from ncrar_audio.babyface import Babyface


device = Babyface('XLR1', 'XLR2', use_osc=False)
cal = FlatCalibration.unity()
waveform = tone(fs=device.fs, duration=1, frequency=1e3, level=-20, calibration=cal)
recording = device.acquire(waveform, 1)
t = np.arange(5000) / device.fs
plt.plot(t, recording[..., :5000].T, label='Recording')

plt.axvline(0.046439909297052155, c='k', label='Sound driver delay')
plt.axvline(0.046439909297052155 + 12 / device.fs, ls=':', c='k',
            label='Sound driver delay plus DA/AD delay')
plt.xlabel('Time (sec)')
plt.ylabel('Amplitude (Volts)')
plt.legend()
plt.show()



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = 'NCRAR audio'
copyright = '2023, NCRAR audio development team'
author = 'NCRAR audio development team'

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    'sphinxcontrib.apidoc',
    'matplotlib.sphinxext.plot_directive',
    'IPython.sphinxext.ipython_directive',
    'IPython.sphinxext.ipython_console_highlighting',
    'sphinx.ext.mathjax',
    'sphinx.ext.autodoc',
    'sphinx.ext.doctest',
    'sphinx.ext.inheritance_diagram',
    'sphinx.ext.napoleon',
    #'sphinx_gallery.gen_gallery',
    'sphinx.ext.autosummary',
]

templates_path = ['_templates']
exclude_patterns = []


apidoc_module_dir = '../../ncrar_audio'
apidoc_output_dir = 'api'
apidoc_separate_modules = True

autodoc_mock_imports = ['pyxid2']
autosummary_generate = True
add_module_names = False


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'pydata_sphinx_theme'
html_static_path = ['_static']



import matplotlib.pyplot as plt
import numpy as np
from scipy import signal

from psiaudio.calibration import FlatCalibration
from psiaudio.stim import ramped_tone
from psiaudio import util

from ncrar_audio.babyface import Babyface


def play_tones():
    calibration = FlatCalibration.unity()

    device = Babyface()
    frequencies = np.array([1e3, 2e3])[:, np.newaxis]
    tones = ramped_tone(device.fs, frequency=frequencies, duration=5,
                       rise_time=0.1, level=-18.94, calibration=calibration)
    recording = device.acquire(tones, input_channels=2)
    recording = signal.detrend(recording)
    psd = util.db(util.psd_df(recording, device.fs))

    # The 1: slice along the columns axis of the dataframe throws out the DC
    # component of the FFT which tends to throw off the Y-axis limits since we
    # detrend the signal.
    psd = psd.iloc[:, 1:]
    plt.plot(psd.columns, psd.iloc[0].values, label='In 1')
    plt.plot(psd.columns, psd.iloc[1].values, label='In 2')

    # The octave scaling is a special one provided by psiaudio. It is only
    # enabled when you import psiaudio in your own code.
    plt.xscale('octave')
    plt.axis(xmin=500, xmax=4000)
    plt.xlabel('Frequency (kHz)')
    plt.ylabel('Signal (dBVrms)')
    plt.show()


if __name__ == '__main__':
    play_tones()



import logging
log = logging.getLogger(__name__)

import socket
import threading
import time

from pythonosc import dispatcher
from pythonosc import osc_server
from pythonosc import udp_client


class OSCClient:
    '''
    Sets up bidirectional communication with a device supporting the Open Sound
    Control protocol.
    '''

    def __init__(self, ip_address=None, send_port=7001, recv_port=9001,
                 warn_on_unhandled=False):
        '''
        Parameters
        ----------
        ip_address : {None, IP address}
            If None, uses the IP address assigned to the local machine.
        send_port : int
            Port to send messages on.
        recv_port : int
            Port to receive messages on.
        warn_on_unhandled : bool
            If True, log all unhandled resposnes from the server to the console
            as a warning. This is usually only for debugging purposes if you
            are trying to figure out what OSC responses you might want to be
            handling.
        '''
        if ip_address is None:
            hostname = socket.gethostname()
            ip_address = socket.gethostbyname(hostname)
        self.send_port = send_port
        self.recv_port = recv_port
        self.ip_address = ip_address
        self.warn_on_unhandled = warn_on_unhandled
        if self.recv_port is not None:
            self._configure_server()
        self._configure_client()

    def _configure_server(self):
        # The dispatcher is responsible for checking whether the datagram
        # (i.e., message recieved on a socket) is in valid OSC format. If it's
        # valid OSC format, then it will dispatch the message to a callback. By
        # default, all messages will get passed to a callback that logs the
        # unhandled message (useful for knowing whether we're "missing out".
        self.dispatch = dispatcher.Dispatcher()
        self.dispatch.map('/', self._heartbeat)
        self.dispatch.set_default_handler(self._log_unhandled_message)

        addr = (self.ip_address, self.recv_port)
        self.server = osc_server.ThreadingOSCUDPServer(addr, self.dispatch)
        self.server_thread = threading.Thread(target=self.server.serve_forever)
        # Be sure to set in daemon so that thread exits cleanly when the master
        # thread exits.
        self.server_thread.daemon = True
        self.server_thread.start()
        log.info('Set up client to recv on %s:%d', *addr)

    def _configure_client(self):
        self.client = udp_client.SimpleUDPClient(self.ip_address,
                                                 self.send_port)
        log.info('Set up client to send on %s:%d', self.ip_address,
                 self.send_port)

    def _heartbeat(self, address, value):
        # This is a heartbeat message that is often transmitted by the OSC
        # server.
        pass

    def _log_unhandled_message(self, *args):
        m = 'Recieved an unhandled OSC response: address=%s, value=%r'
        log.warning(m, *args)

    def send_message(self, address, value, response_address=None):
        if response_address is not None:
            # TODO: We need to figure out how to attach the response to the
            # specific request. asyncio is perfect for this type of dispatch.
            raise NotImplementedError
        log.info('Sending OSC command: address=%s, value=%r', address, value)
        self.client.send_message(address, value)

    def send_messages(self, messages):
        '''
        Send a list of messages

        Each message must be a two or three element iterable. The following are
        valid message formats:

            ('/1/volume1', 1.0, '/1/volume1'),
            ('/1/busOutput', 1),
        '''
        for message in messages:
            self.send_message(*message)
            time.sleep(0.001)


if __name__ == '__main__':
    # Set up basic logging config so we can see the logging messages printed
    # out
    logging.basicConfig(level='INFO')

    device = OSCClient()
    osc_messages = [
        ["/1/busPlayback", 1.0],
        ["/1/busInput", 1.0],
        ["/1/busOutput", 1.0],
        ["/1/volume1",0.7],
        ["/1/volume2",0.7],
        ["/1/volume3",0.7],
        ["/1/volume4",0.7],
        ["/1/volume5",0.7],
        ["/1/volume6",0.7],
        ["/1/volume7",0.7],
        ["/1/volume8",0.7],
        ["/1/volume9",0.7],
        ["/1/volume10",0.7],
        ["/1/volume11",0.7],
        ["/1/volume12",0.7],
    ]
    device.send_messages(osc_messages)
    time.sleep(1)



import time

from ncrar_audio.babyface import Babyface


def main():
    device = Babyface()
    device.set_volume(-6)
    time.sleep(0.1)
    print(device._volume_db)

    #messages = []
    #for i in range(12):
    #    messages.append(['/setSubmix', i])
    #    messages.append(['/1/busPla', i])

    submix = 1
    print(f'Set submix {submix}')
    device.osc_client.send_messages([['/setSubmix', submix]])
    for channel in range(12):
        device.osc_client.send_messages([['/1/busOutput', 1]])
        device.osc_client.send_messages([[f'/1/volume{channel+1}', 0]])
        device.osc_client.send_messages([['/1/busPlayback', 1]])
        device.osc_client.send_messages([[f'/1/volume{channel+1}', 0]])
        time.sleep(0.5)

    return


    messages = []
    for submix in range(12):
        messages.extend([
            ['/busOutput', 1.0],
            [f'/1/volume{submix+1}', 0.818],
            ["/1/busPlayback", 1.0],
        ])
        for channel in range(12):
            volume = 0.818 if channel == submix else 0
            messages.append([f"/1/volume{channel+1}", volume])

    print(messages)

    #set_submix = [
    #        ["/setSubmix", 0.0],
    #        ["/1/volume1",0.818],
    #        ["/1/busOutput", 1.0],
    #        ["/1/volume1",0.818],
    #        #["/setsubmix", 2.0],
    #        #["/1/busOutput", 1.0],
    #        #["/1/volume3",0.818],
    #]
    #device.osc_client.send_messages(messages)
    #device.send_messages([
    #    ['/1/setSubmix', 0],
    #    ['1/busPlayback', 1],
    #    ['1/busPlayback', 1],



if __name__ == '__main__':
    main()



from .version import __version__

import os


def disable_quick_edit():
    # From https://stackoverflow.com/questions/73486528/python-script-pausing-in-cmd
    import win32console as con
    import signal

    # Missing constants in pywin
    ENABLE_EXTENDED_FLAGS = 0x0080
    ENABLE_QUICK_EDIT_MODE = 0x0040

    # Modify console mode to disable quick edit mode
    h = con.GetStdHandle(con.STD_INPUT_HANDLE)
    oldMode = h.GetConsoleMode()
    h.SetConsoleMode((oldMode | ENABLE_EXTENDED_FLAGS) &
            ~ENABLE_QUICK_EDIT_MODE)


# This ensures that running scripts from the command line does not accidentally
# pause the script (such as for tqdm callbacks).
if os.name == 'nt':
    try:
        disable_quick_edit()
    except:
        pass






def test_ramped_tone():
    import logging
    logging.basicConfig(level='INFO')

    import matplotlib.pyplot as plt
    from scipy import signal
    import numpy as np

    from psiaudio.api import ramped_tone, FlatCalibration
    calibration = FlatCalibration.unity()

    device = Babyface()
    frequencies = np.array([1e3, 2e3])[:, np.newaxis]
    tone = ramped_tone(device.fs, frequency=frequencies, duration=15,
                       rise_time=0.1, level=-16.7, calibration=calibration)
    print(tone.max(axis=1))
    device.play(tone)
    '''
    #recording = device.record(int(device.fs)*2, 2)
    recording = signal.detrend(recording)

    print(np.abs(recording).max(axis=-1))
    from psiaudio import util
    print(util.rms(recording))

    psd = util.db(util.psd_df(recording, device.fs))
    print(psd.max().max())

    plt.plot(psd.columns, psd.iloc[0].values, label='In 1')
    plt.plot(psd.columns, psd.iloc[1].values, label='In 2')
    plt.xlabel('Frequency (Hz)')
    plt.ylabel('Signal (dBVrms)')
    plt.show()
    '''

def play_cal_tone():
    import pickle
    import numpy as np
    from psiaudio.api import ramped_tone, FlatCalibration
    from psiaudio import util
    calibration = FlatCalibration.unity()

    device = Babyface()
    frequencies = np.array([1e3])[:, np.newaxis]
    tone = ramped_tone(device.fs, frequency=frequencies, duration=15,
                       rise_time=0.1, level=-18.94, calibration=calibration)
    recording = device.record(tone.shape[-1], 2)
    print(recording.shape)
    print(util.rms(recording))
    np.savetxt('ER7c calibration.txt', recording)
    with open('ER7c calibration.pkl', 'wb') as fh:
        pickle.dump(recording, fh)


def test_volume_control():
    mapping = load_volume_map()
    print(mapping(0))


if __name__ == '__main__':
    test_volume_control()
    play_cal_tone()




import logging
log = logging.getLogger(__name__)

from pathlib import Path
import struct

import numpy as np
import pandas as pd
from scipy import signal

from psiaudio import util

# Note, most of this code is derived from code written by Sam Gordon. I don't
# have the original reference documentation from IHS (despite asking for it),
# so I've done my best. The important thing is that the epochs are properly
# returned!

def read_cam(filename, header_only=False):
    description = [
        ('std_ihs_ep', 1000, 'h'),
        ('cont_data_suppl', 4000, 'i'),
        ('seq1_locs', 5000, 'i'),
        ('seq1_align', 5000, 'i'),
        ('seq2_locs', 5000, 'i'),
        ('seq2_align', 5000, 'i'),
        ('marker_locc', 4000, 'i'),
        ('marker_types', 4000, 'i'),
        ('additional_dloc', 12500, 'i'),
        ('short_string', None, f'{125*32}s'),
        ('remainder_string', None, '16000s'),
    ]

    recording_type_map = {
        0: 'raw',
        1: 'std average',
        2: 'clad',
        3: 'sequential stimulation',
        4: 'complex sequence average'
    }


    with Path(filename).open('rb') as fh:
        fh.seek(0, 2)
        fh_nbytes = fh.tell()
        fh.seek(0, 0)

        result = {}
        for (name, count, dtype) in description:
            if count is None:
                nbytes = struct.calcsize(dtype)
                value = struct.unpack(dtype, fh.read(nbytes))[0]
                if dtype.endswith('s'):
                    value = value.decode('utf-8').replace('\x00', ' ')
                result[name] = value
            else:
                result[name] = np.fromfile(fh, dtype=dtype, count=count)

        n_channels = result['n_channels'] = result['cont_data_suppl'][0]

        if not header_only:
            # There are four channels. the int16 (2-byte) samples are stored in
            # interleaved fashion.
            fh.seek(200000, 0)
            recording = np.fromfile(fh, dtype='h').reshape((-1, n_channels)).T
            result['dio'] = recording[-1]
            result['recording'] = recording[:-1]

    result['sample_time'] = result['cont_data_suppl'][6] / 1e9
    result['fs'] = 1 / result['sample_time']

    recording_type = result['cont_data_suppl'][1]
    result['recording_type'] = recording_type_map[recording_type]

    # 1 rare, 2 cond, 3 alt
    result['recording_phase'] = result['cont_data_suppl'][2]

    # 0 off, 1 on
    result['digital_channel'] = result['cont_data_suppl'][3]

    result['n_sequences'] = result['cont_data_suppl'][4]
    result['n_unique_stimuli'] = result['cont_data_suppl'][5]
    result['seq_sweep_length'] = result['cont_data_suppl'][249]
    result['seq_sweep_amount'] = result['cont_data_suppl'][250]
    result['n_stimuli'] = result['cont_data_suppl'][251]
    result['stim_rate'] = result['cont_data_suppl'][252] / 1000
    result['seq_ear_send'] = result['cont_data_suppl'][253]

    # 0 alt, 2 cond, 1 rare, 3 complex
    result['seq_stim_mode'] = result['cont_data_suppl'][254]
    result['seq_stim_send'] = result['cont_data_suppl'][255]

    result['stim_settings'] = stim_settings = []
    for i in range(result['n_stimuli']):
        stim_settings.append({
            # intensity in db
            'intensity': result['cont_data_suppl'][310 + i * 15],
            # 1 is click, 2 is toneburst
            'type': result['cont_data_suppl'][311 + i * 15],
            # frequency of tone pip, 0 if click
            'frequency': result['cont_data_suppl'][312 + i * 15],
            # duration of stimulus in number of points
            'duration': result['cont_data_suppl'][313 + i * 15],
            # calibration value in db
            'right_calibration': result['cont_data_suppl'][314 + i * 15],
            # calibration value in db
            'left_calibration': result['cont_data_suppl'][315 + i * 15],
            # hl = 0, spl = 1
            'hlspl_flag': result['cont_data_suppl'][316 + i * 15],
            # rise/fall time (in points) of stim
            'stim_rise_fall_time': result['cont_data_suppl'][317 + i * 15],
            # stimulator type ???
            'stimulator_type': result['cont_data_suppl'][318 + i * 15],
        })

    # Apparently channel #4 is the digital channel
    channel_settings = []
    for i in range(result['n_channels']):
        channel_settings.append({
            # gain in uv?
            'gain': result['cont_data_suppl'][400 + i * 10] / 100,
            # what units???
            'high_pass': result['cont_data_suppl'][401 + i * 10] / 1000,
            'low_pass': result['cont_data_suppl'][402 + i * 10] / 1000,
            # notch filter on or off
            'notch': result['cont_data_suppl'][403 + i * 10] / 1000,
        })
    result['channel_settings'] = pd.DataFrame(channel_settings)
    return result


def extract_timestamps(dio):
    ts = {}

    # These are the start and end markers of the sequence. In at least one file
    # I looked at, the end marker was missing for some reason. We need to
    # handle that edge-case.
    m_start_end = (dio > -30e3) & (dio <= -20e3)
    bounds = np.flatnonzero(m_start_end)
    bounds = [int(b) for b in bounds]
    if len(bounds) == 2:
        lb, _ = ts['start'], ts['end'] = bounds
    elif len(bounds) == 1:
        # If end marker is missing, then set it to None
        lb, _ = ts['start'], ts['end'] = bounds[0], None
    elif len(bounds) == 0:
        # No start marker found for the DIO
        lb = None

    ts['stim'] = {}
    get_id = lambda x: int(np.abs(x) % 1000)

    # Check to see if start marker of sequence also encodes start marker of
    # stimulus.
    if get_id(lb) != 0:
        ts['stim'].setdefault(get_id(dio[lb]), []).append(lb)

    # Now, pull out the stimulus start times.
    m_stim = (dio > -17e3) & (dio <= -16e3)
    for i in np.flatnonzero(m_stim):
        ts['stim'].setdefault(get_id(dio[i]), []).append(i)
    for k, v in ts['stim'].items():
        ts['stim'][k] = np.array(v)

    m = dio < -32000
    ts['errors'] = {
        'n_missing_timepoints': int(m.sum()),
        'missing_segments': util.epochs(m),
    }

    return ts


def extract_cam_epochs(result, offset=0, duration=0.5, filter_lb=80,
                       filter_ub=3000, filter_order=800, ts=None):

    # Find the timestamps we want to extract.
    if ts is None:
        ts = extract_timestamps(result['dio'])

    # First, filter the signal so we don't have to worry about padding to
    # control for filter edges when epoching.
    if filter_order is not None:
        b = signal.firwin(filter_order, (filter_lb, filter_ub),
                          fs=result['fs'], pass_zero='bandpass')
        w = signal.filtfilt(b, 1, result['recording'], axis=-1)
    else:
        w = result['recording']
    dio = result['dio']

    # Find the start/end edges of the epochs
    o = int(round(result['fs'] * offset))
    d = int(round(result['fs'] * duration))

    all_epochs = {}
    n_error = 0

    missing = ts['errors']['missing_stim_segments'] = {}
    for key, indices in ts['stim'].items():
        missing[key] = []
        epochs = []
        i_start = indices + o
        i_end = i_start + d

        for i, (lb, ub) in enumerate(zip(i_start, i_end)):
            m = dio[lb:ub] < -32000
            if m.any():
                n_error += 1
                missing[key].append(i)
            else:
                e = w[..., lb:ub][np.newaxis]
                if e.shape[-1] != d:
                    log.warn('Dropping truncated epoch %d', i)
                    missing.setdefault(key, []).append(i)
                else:
                    epochs.append(e)

        log.warn('Found %d epochs for stim ID %d', len(epochs), key)
        all_epochs[key] = np.concatenate(epochs)

    if n_error:
        log.warn('Could not read %d epochs due to dropped buffer', n_error)
    return all_epochs



import pytest

import numpy as np

from ncrar_audio import triggers


@pytest.fixture(scope='module', params=[512, 16384])
def fs(request):
    return request.param


@pytest.fixture(scope='module', params=['cos', 'square'])
def trigger_shape(request):
    return request.param


@pytest.fixture
def n_unique_trigs():
    return 3


@pytest.fixture
def n_trigs(n_unique_trigs):
    # Must be a multiple of the number of unique trigger groups
    return 4 * n_unique_trigs


@pytest.fixture
def trig_iti():
    return 0.1


@pytest.fixture
def trigger_train(fs, trigger_shape, n_unique_trigs, n_trigs, trig_iti=0.1):
    n_samples = int(fs * trig_iti)

    trigs = []
    for repeat in range(n_unique_trigs):
        t = triggers.make_trigger(fs, n_samples, shape=trigger_shape, shape_settings={'repeat': repeat+1})
        trigs.append(t)

    t = np.concatenate(trigs, axis=-1)
    t = np.tile(t, int(n_trigs / n_unique_trigs))
    t = np.pad(t, 10, 'constant')
    return t


def test_extract_triggers(fs, n_unique_trigs, n_trigs, trig_iti, trigger_train):
    group_window = int(0.025 * fs)
    trigs = triggers.extract_triggers(trigger_train, group_window=group_window)
    assert len(trigs) == n_unique_trigs
    for code in range(n_unique_trigs):
        assert len(trigs[code + 1]) == n_trigs / n_unique_trigs
        iti = np.mean(np.diff(trigs[code + 1]) / fs)
        assert pytest.approx(iti, abs=1e-2) == (trig_iti * n_unique_trigs)



from pathlib import Path
import struct

import numpy as np


general_header = [
    ('stim_type', 'i'),
    ('description', '200s'),
    ('cal_info', '20s'),
    ('fs', 'd'),
    ('duration', 'i'),
    ('max_amplitude', 'i'),
    ('component_count', 'i'),
    ('zero_position', 'i'),
    ('_reserved', '200b')
]


component_header = [
    ('component_type', 'i'),
    ('frequency', 'd'),
    ('phase', 'd'),
    ('delay', 'i'),
    ('duration', 'i'),
    ('window', 'i'),
    ('rise_fall_time', 'i'),
    ('per_amp', 'd'),
    ('resp_freq', 'd'),
    ('SPL_to_HL', 'i'),
    ('_reserved', '36b'),
    ('active', 'b'),
    ('amp_mod', 'b'),
    ('freq_mod', 'b'),
    ('resp_seq', 'b'),
    ('calculate', 'b'),
    ('_reserved', '9b'),
    ('amp_mod_freq', 'd'),
    ('amp_mod_per', 'd'),
    ('amp_mod_phase', 'd'),
    ('_reserved', '40b'),
    ('freq_mod_freq', 'd'),
    ('freq_mod_per', 'd'),
    ('freq_mod_phase', 'd'),
    ('_reserved', '40b'),
    ('rep_seq_duration', 'i'),
    ('rep_seq_count', 'i'),
    ('rep_seq_freq', 'd'),
    ('rep_seq_isi', 'i'),
    ('rep_seq_ramp_level', 'i'),
    ('group_delay', 'i'),
    ('rep_seq_freq_range', 'd'),
    ('_reserved', '24b'),
    ('_reserved', '200b'),
]


cal_description = [
    ('right_cal', 'h'),
    ('right_max', 'h'),
    ('right_min', 'h'),
    ('left_cal', 'h'),
    ('left_max', 'h'),
    ('left_min', 'h'),
    ('spl_conversion', 'h'),
    ('offset_cal_flag', 'h'),
    ('ref_freq', 'i'),
    ('right_offset', 'h'),
    ('left_offset', 'h'),
    ('_reserved', '190b'),
]


def _read_description(description, fh):
    keys, fmt = zip(*description)
    descr = '=' + ''.join(fmt)
    nbytes = struct.calcsize(descr)
    values = struct.unpack(descr, fh.read(nbytes))

    result = {}
    for key, typecode, value in zip(keys, fmt, values):
        if key.startswith('_'):
            continue
        if typecode.endswith('s'):
            value = value.decode('utf-8').lstrip('\r').rstrip('\x00')
        result[key] = value
    return result


def read_stm(filename):
    with Path(filename).open('rb') as fh:
        # Get number of bytes in file. This is a sanity-check.
        fh.seek(0, 2)
        fh_nbytes = fh.tell()
        fh.seek(0, 0)

        b_version = fh.read(2)
        version, = struct.unpack('h', b_version)
        if version != 310:
            raise ValueError('Unsupported version')

        general = _read_description(general_header, fh)
        components = {}
        for i in range(general['component_count']):
            components[i] = _read_description(component_header, fh)

        caldata = {}
        for i in range(50):
            caldata[i] = _read_description(cal_description, fh)

        waveform_start = fh.tell()
        b_waveform = fh.read(general['duration'] * 2)
        waveform_end = fh.tell()

        waveform = np.frombuffer(b_waveform, 'h')

        # Make sure the file was parsed properly
        if len(waveform) != general['duration']:
            raise ValueError('Data missing')

        remaining = fh_nbytes - fh.tell()
        if remaining > 0:
            raise ValueError(f'File has {remaining} byes of unread data')

    return {
        'version': version,
        'header': general,
        'component_headers': components,
        'calibration_data': caldata,
        'waveform': waveform,
        'file_stats': {
            'waveform_start': waveform_start,
            'waveform_end': waveform_end,
        },
    }



import matplotlib.pyplot as plt
import numpy as np

from psiaudio.calibration import FlatCalibration
from psiaudio.stim import tone
from ncrar_audio.babyface import Babyface


device = Babyface('XLR1', None, use_osc=False)
cal = FlatCalibration.unity()
waveform = tone(fs=device.fs, duration=1, frequency=1e3, level=-20, calibration=cal)

# This allows you to start playing audio and then do other operations in a
# non-blocking fashion.
with device.play_async(waveform) as stream:
    print('Async stream started')
    while not device.wait(timeout=0.1):
        print('... still waiting')
    print('Async stream done!')


# This just plays the sound and then exits once it's done.
print('Stream started')
device.play(waveform)
print('Stream done!')
